<?php 

 /* amavis-stats.php - build rrd graphs from amavis-stats collected data.

 Copyright (C) 2003,2004 Mark Lawrence (nomad@null.net) -->

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License with
 the Debian GNU/Linux distribution in file /usr/share/common-licenses/GPL;
 if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 Suite 330, Boston, MA  02111-1307  USA
 
 On Debian systems, the complete text of the GNU General Public
 License, version 2, can be found in /usr/share/common-licenses/GPL-2. */


/*
*  Setup the start of the html
*/
?>
<!DOCTYPE HTML PUBLIC \-//W3C//DTD HTML 3.2 Final//EN\>
<html>
<head>

<style>
    h1 {
        font-size: 1.2em;
        float: left;
    }
    body {
        font-size: small;
        font-family: sans-serif; 
        background-color: #e0e0e0;
        margin: 1em;
    }
    img {
        margin-bottom: 1em;
        text-align: center;
        clear: both;
    }
    a {
        text-decoration: none;
    }
    a:hover {
        text-decoration: underline;
        color: #000000;
    }
    .normal {
        font-size: .8em;
    }
    .right {
        float: right;
    }
    .left {
        float: left;
    }
    input {
        font-size: x-small;
    }
</style>

<?php

/*
*
* The rrdstep value MUST match that which is used by the perl script.
*/
$rrdstep = 300;

/*
* php version of the RRDTool::rrd_graph extention using shell executable.
* This needs to come first in the file otherwise the php pre-compiler
* thinks we don't exist...
*/

$rrd = "shared-library";

if (!function_exists('rrd_graph') && !@dl('rrdtool.so')) {
    $rrd = "command-line";

    function rrd_graph ($img, $opts, $count) {
        global $rrd_error_val;

        unset($output);

        $cmd = "rrdtool graph $img '". implode("' \\\n'",$opts) . "'" .
               " 2>&1";
        aslog($cmd, 1);
        $out = exec($cmd, $output, $rrd_error_val);

        if ($rrd_error_val) {
            if (!is_numeric($rrd_error_val)) {
                $rrd_error_val = $output[0];
            }
            return false;
        }

        aslog($output);

        $retval = array();

        if (preg_match("/(\d+)x(\d+)/", $output[0], $matches)) {
            $retval["xsize"] = $matches[1];
            $retval["ysize"] = $matches[2];
        } else {
            $rrd_error_val = $output[0];
            return false;
        }

        array_shift($output);

        $retval["calcpr"] = array();
        foreach ($output as $index => $value) {
            $retval["calcpr"][$index] = $value;
        }
        return $retval;
    }

    function rrd_error() {
        global $rrd_error_val;
        return $rrd_error_val;
    }
}


/*
* Debugging function
*/
function pre_r($obj) {
    print "<pre>\n";
    print_r($obj);
    print "</pre>\n";
}


/*
* Timing function to work out how long things take
*/
function elapsed($start)
{
    $end = microtime();
    list($start2, $start1) = explode(" ", $start);
    list($end2, $end1) = explode(" ", $end);
    $diff1 = $end1 - $start1;
    $diff2 = $end2 - $start2;
    if( $diff2 < 0 ){
        $diff1 -= 1;
        $diff2 += 1.0;
    }
    return $diff2 + $diff1;
}


/*
* Get and set the defaults for various options
*/
$asVersion   = "0.1.13-rc6";
$me          = $HTTP_SERVER_VARS["SCRIPT_NAME"];
$libdir   = "/var/lib/amavis-stats";
$namefile = $libdir."/amavis-stats.names";
$imgdir      = "img";

$now = time();
if (isset($_GET["time"])) {
    $now = $_GET["time"];
}
    
$rate = 3600;
if (isset($_GET["rate"])) {
    $rate = $_GET["rate"];
}

$debug = false;
if (isset($_GET["debug"])) {
    $debug = true;
}
 
$host = "";
if (isset($_GET["host"])) {
    $host      = $_GET["host"];
    $libdir = $libdir."/$host";
    $imgdir    = "img/$host";
}

if (!is_dir($imgdir)) {
    if (!mkdir($imgdir)) {
        asErr("Could not create $imgdir");
    }
}
 
$refresh = 0;

if (isset($_GET["refresh"]) && 
    is_numeric($_GET["refresh"]) && 
    $_GET["refresh"] > 0) {
        $refresh = $_GET["refresh"];
        print "<meta http-equiv=\"refresh\" content=\"$refresh;\">";
}

$span = "daily";
if (isset($_GET["span"])) {
    $span = $_GET["span"];
}


/*
* Build the title with links to each host
*/
print "
</head>
<body>
<h1>Amavis Detection Statistics
";

$path = "/var/lib/amavis-stats";
$handle=opendir($path);

$hosts = array();
while ($file = readdir($handle)) {
    if ($file != "." && is_dir("$path/$file")) {
        if (file_exists("$path/$file/amavis-stats.state")) {
            $hosts[] = $file;
        }
    }
}

if (file_exists("$path/amavis-stats.state") && (count($hosts) == 0)) {
    print "for ".eregi_replace("\n", "", `hostname`);
} else {
    $separator = "for";
    foreach($hosts as $target) {
        if (strlen($target) == 0) {
            continue;
        } elseif ($target != $host) {
            print "$separator
               <a href=\"$me?host=$target&span=$span&refresh=$refresh\">
               $target</a>\n";
        } else {
            print "$separator $target ";
        }
        $separator = "|";
    }

    if (file_exists("$path/amavis-stats.state")) {
        if (strlen($host) == 0) {
            print "$separator ALL";
        } else {
         print "| <a href=\"$me?span=$span&refresh=$refresh\">ALL</a>\n";
        }
    }
}


/*
* Build the daily, weekly, monthly, yearly menu on the right hand side.
* Also build the "refresh" form for regular reloads of the page
*/
print "
</h1>
<div class=\"right\">
";

if ($span != "daily") {
    print "<a href=\"$me?host=$host&span=daily&refresh=$refresh\">daily</a> | ";
} else {
    print "daily | ";
}
if ($span != "weekly") {
    print "<a href=\"$me?host=$host&span=weekly&refresh=$refresh\">weekly</a> | ";
} else {
    print "weekly | ";
}
if ($span != "monthly") {
    print "<a href=\"$me?host=$host&span=monthly&refresh=$refresh\">monthly</a> | ";
} else {
    print "monthly | ";
}
if ($span != "yearly") {
    print "<a href=\"$me?host=$host&span=yearly&refresh=$refresh\">yearly</a>";
} else {
    print "yearly";
}
?>


<br>
<br>

<form action="<?php print "$me"; ?>">Refresh after
    <input name="refresh" size="4" value="<?php print $refresh; ?>"> seconds.
    <input name="host" type="hidden" value="<?php print $host; ?>"> 
    <input name="span" type="hidden" value="<?php print $span; ?>">
</form>

</div>
<br style="clear: left;">
<?php


/*
*  Load the statistics and call the graph generating functions
*/
if (!asLoadStats($host)) {
    asMsg("No statistics available.");
} else {
    
    print asPGraph("$imgdir/passed-$span.png", $now, $span);
    print asVGraph("$imgdir/virus-$span.png", $now, $span);
}


/*
*  Finish up the html with the description of the message classifications
<p>

<span style="font-size:x-small;">
<b>Passed</b>:        clean or $tag_level &lt; $tag2_level<br>
<b>Passed(SPAM)</b>:  classified as SPAM but delivered because
                      $tag_level &lt; $kill_level<br>
<b>Banned</b>:        not delivered, e.g. banned file extension attached<br>
<b>Infected</b>:      email contains virus<br>
<b>Not Delivered</b>: could not be delivered, e.g. destination host 
                      refuses to accept delivery<br>
<b>Not-Delivered(SPAM)</b>: SPAM classified with $tag_level &gt;=
                       $kill_level and $final_spam_destiny != pass<br>
*/


if ($refresh != 0 && $refresh < $rrdstep) {
    print "<p>Warning: Your refresh rate is less than the rrdstep size of the RRDs. You will be refreshing without any new data to see.\n";
}
print "<br>Generated by <a
    href=\"http://rekudos.net/amavis-stats/\">amavis-stats</a>
    version $asVersion using $rrd rrdtool ";

if (!$debug) {
    print "[<a href=\"$me?host=$host&span=$span&refresh=$refresh&debug=1\">debug</a>";
} else {
    print "[<a href=\"$me?host=$host&span=$span&refresh=$refresh\">nodebug</a>";
}
print "]\n </span> </body> </html>";


/*
*
*/
function aslog($txt = "", $pre = 0) {
    if ($GLOBALS["debug"]) {
        if ($pre) {
            print "<pre>\n";
        }
        print "amavis-stats::debug: $txt\n";
        if (is_array($txt)) {
            print "<pre>\n";
            print_r($txt);
            print "</pre>\n";
        }
        if ($pre) {
            print "</pre>\n";
        }
        print "<br>\n";
    
    }
}

function asMsg($txt = "") {
    print "amavis-stats: $txt<br>\n";
}

function asErr($txt = "") {
    print "amavis-stats::error: $txt<br>\n";
}






/*
*
*/
function asLoadStats ($host) {
    global $libdir, $namefile, $virus, $maxi, $msgs;

    $statefile = "$libdir/amavis-stats.state";
    $seenfile  = "$libdir/amavis-stats.seen";


    $readfile = @file($seenfile);
    if (!is_array($readfile)) {
        asErr("Could not open $seenfile");
        return false;
    }
    aslog("$seenfile");

    $virus = array();
    for ($k = 0; $k <= (count($readfile) - 1); $k++) {
        $fields = preg_split("/\s+/",$readfile[$k], -1, PREG_SPLIT_NO_EMPTY);
        $virus[$fields[0]]["id"] = $fields[0];
        $virus[$fields[0]]["firstseen"] = $fields[1];
        $virus[$fields[0]]["lastseen"] = $fields[2];
    }

    $readfile = @file($namefile);
    if (!is_array($readfile)) {
        asErr("Couldn't open id => name mappings file.");
    }
    aslog("$namefile");

    $msgs = array();
    for ($k = 0; $k <= (count($readfile) - 1); $k++) {
        $fields = preg_split("/\s+/",$readfile[$k], 2, PREG_SPLIT_NO_EMPTY);

        $id = trim($fields[0]);
        $name = trim($fields[1]);

        if (isset($virus[$id])) { /* ID has been seen */
            $virus[$id]["name"] = $name;  /* so record the name */
        } else {
            continue; /* otherwise we don't need to know about this guy */
        }

        /*
        * Pull out the "standard" IDs (Passed *, Blocked *)
        */
        if (strpos($name, "Passed") === 0 ||
            strpos($name, "Blocked") === 0) {
            $msgs[$id] = $name;
            unset($virus[$id]);
            continue;
        }

        $vid[$name] = $id;

    }

    aslog($virus);
    aslog($vid);
    aslog($msgs);

    if (count($virus) == 0 and count($msgs) == 0) {
        asErr("No viruses/IDs seen.");
        return false;
    }

    $readfile = @file($statefile);
    if (!is_array($readfile)) {
        asErr("Couldn't open state file.");
    }
    aslog($statefile);
    aslog($readfile);

    for ($k = 0; $k <= (count($readfile) - 1); $k++) {
        $fields = preg_split("/\s+/",$readfile[$k], -1, PREG_SPLIT_NO_EMPTY);
        if ($fields[0] == "lastupdate:" && is_numeric($fields[1])) {
            $lastupdate = $fields[1]; 
        } elseif ($fields[0] == "LC_TIME:") {
            setlocale(LC_TIME, $fields[1]);
        }
    }

    if (!isset($lastupdate)) {
        asErr("lastupdate not defined.");
        return false;
    }
    elseif ($lastupdate == 0) {
        asErr("last update was at 0 seconds.");
        return false;
    }

    return true;
}


/*
* The beginning and the end of the calculations must not fall half
* way through any of the rrd "bins", but only on the boundaries. Which
* bin is used is determined by rrdtool based on the span.
*
* The rrds are created as follows (from the amavis-stats perl script)
* with a "step" of 300:
*
*   "RRA:AVERAGE:0.5:1:300"   = 1 bin every step (300 seconds)
*   "RRA:AVERAGE:0.5:6:700"   = 1 bin every 6 steps (1800 seconds - 1/2 hour)
*   "RRA:AVERAGE:0.5:24:775"  = 1 bin every 24 steps (7200 seconds - 2 hours)
*   "RRA:AVERAGE:0.5:288:797" = 1 bin every 288 steps (86400 seconds - 1 day)
*
* If we make sure that for the appropriate span we start and end on the
* appropriate place the totals should work out accurately.
*
* By the same logic, we have to make sure that the rrd files are first
* populated with data at the beginning of the highest bin (ie at the start
* of a day). Otherwise we are calculating over a period which contains
* some *UNKNOWN* data.
*
* At least, that's the theory.
*
* getlimits($end, $span)
*
* $end  - end time in seconds
* $span - string either "daily", "weekly", "monthly", or "yearly"
*
* Returns an array with the new end, a start value, and a better text
* description of the span.
*/
function getlimits($end, $span) {
    global $rrdstep;

    /* how wide is each bin, in seconds? */
    $dayw     = $rrdstep;
    $weekw    = 6*$rrdstep;
    $monthw   = 24*$rrdstep;
    $yearw    = 288*$rrdstep;

    /* how long is each span in seconds? */
    $daysec   = 60*60*24;
    $weeksec  = $daysec*7;
    $monthsec = $weeksec*4;
    $yearsec  = $weeksec*52;

    /* make sure that the span is a multiple of the bin size */
    /* is this actually necessary? */
    $dayspan   = floor($daysec/$dayw) * $dayw;
    $weekspan  = floor($weeksec/$weekw) * $weekw;
    $monthspan = floor($monthsec/$monthw) * $monthw;
    $yearspan  = (int)($yearsec/$yearw) * $yearw;

    $result   = array();

    if ($span == "daily") {
        $result["end"]         = floor($end/$dayw) * $dayw;
        $result["start"]       = $result["end"] - $dayspan;
        $result["description"] = "24 hours";
    } elseif ($span == "weekly") {
        $result["end"]         = floor($end/$weekw) * $weekw;
        $result["start"]       = $result["end"] - $weekspan;
        $result["description"] = "1 week";
    } elseif ($span == "monthly") {
        $result["end"]         = floor($end/$monthw) * $monthw;
        $result["start"]       = $result["end"] - $monthspan;
        $result["description"] = "4 weeks";
    } elseif ($span == "yearly") {
        $result["end"]         = floor($end/$yearw) * $yearw;
        $result["start"]       = $result["end"] - $yearspan;
        $result["description"] = "52 weeks";
    } else {
        asErr("getlimits: bad span: $span");
    }

    aslog($result);
    return $result;
}


/*
*
*/
function addopts(&$opts, $type, $id, $vcount, $virus, $length) {
    global $libdir, $as_colors, $rate, $maxi, $host;

    $col  = substr(md5($virus[$id]["name"]),7,6);
    $name = sprintf("%-".$maxi."s", $virus[$id]["name"]);
    $count = $vcount[$id];
    $count = sprintf("%8d", $count);
    $opts[] = "DEF:v$id=$libdir/$id.rrd:hits:AVERAGE";
//    $opts[] = "CDEF:gv$id=v$id,$rate,*";
    $opts[] = "CDEF:gv$id=v$id,UN,0,v$id,IF,$rate,*";
    $opts[] = "$type:gv$id#$col:$name $count";

    return $opts;
}


/*
* asPGraph($img, $end, $span)
*
* $img  - name of the png to generate
* $end  - end time in seconds
* $span - string either "daily", "weekly", "monthly", or "yearly"
*
* Build a graph of clean or "Passed" emails.
* Returns either a html-valid <img> tag which can be printed, or the
* boolean "false" if something went wrong.
*/
function asPGraph($img, $end, $span) {
    global $me, $libdir, $virus; 
    global $maxi, $rate, $asVersion, $rrdstep, $host, $msgs;

    /*
    * Work out the beginning and the end, based on the span
    */
    $limits   = getlimits($end, $span);
    $end      = $limits["end"];
    $start    = $limits["start"];
    $length   = $end - $start;
    $timetext = $limits["description"];

    $startdate = strftime("%c", $start);
    $enddate   = strftime("%c %Z", $end);
    $nowdate   = strftime("%c", time());

    $hostname = "on $host ";
    if ($host == "") {
        $hostname = "";
    }

    if ($rate == 3600) {
        $ratemsg = "hour";
    } else {
        $rate = 60;
        $ratemsg = "min";
    }

    $arrcol = array("Passed SPAM"      => "6a5acd",
                    "Blocked SPAM"     => "fa5acd",
                    "Passed BANNED"    => "cc9900",
                    "Blocked BANNED"   => "ffdd00",
                    "Passed INFECTED"  => "9900aa",
                    "Blocked INFECTED" => "ff3a3d",
                    "Passed CLEAN"     => "10d0d0",
                    "Blocked CLEAN"    => "ff3add");
    

    /*
    * It is a two-step process to build the final graph. The average over
    * a specific time period seems to be impossible to get without actually
    * building a graph. Ie, rrd fetch will not calculate the values we
    * need - we would have to sum and average manually.
    *
    * However the PRINT function of a graph will return what we want
    * in an array. So first of all build a graph that PRINTs the average
    * of every virus over the selected time period.
    */

    $opts = array();
    $opts[] = "--start=$start";
    $opts[] = "--end=$end";

    foreach ($msgs as $id => $name) {
        $opts[] = "DEF:v$id=$libdir/$id.rrd:hits:AVERAGE";
        $opts[] = "CDEF:gv$id=v$id,UN,0,v$id,IF";
        $opts[] = "CDEF:gvt$id=gv$id,$length,*";
        $opts[] = "PRINT:gvt$id:AVERAGE:%.0lf";
    }

    $ret = rrd_graph($img, $opts, count($opts));

    /*
    * debugging - graph definitions
    */
    aslog($ret);

    if (!is_array($ret)) {
        $msg = rrd_error();
        asErr("rrd_graph(): $msg");
        return false;
    }

    /*
    * All results from PRINT commands are in the array $ret["calcpr"][..]
    */
	$maxi = 0;
    $i = 0;
    $vcount = array();
    foreach ($msgs as $id => $name) {
        /*
        * We don't have enough resolution in the rrds
        * to calculate the correct counts at low averages,
        * so we just don't display them
        */
        if ($ret["calcpr"][$i] != 0) {
            $vcount[$id] = $ret["calcpr"][$i];
            $maxi = max($maxi, strlen($name));
        }
        $i++;
        /* aslog("Id: $id = $vcount[$id]"); */
    }
    $maxi++;
    aslog("Maxi: $maxi");

    /*
    */
    if (count($vcount) >= 1) {
        arsort($vcount);
    }
    else {
        aslog("vcount is an empty array");
    }
    aslog($vcount);

    $opts = array();
    $opts[] = "--start=$start";
    $opts[] = "--end=$end";
    $opts[] = "--imgformat=PNG";
    $opts[] = "--title=Message Breakdown $hostname($timetext to $enddate)";
    $opts[] = "--width=520";
    $opts[] = "--vertical-label=msgs/$ratemsg";

    $type = "AREA";

    foreach ($vcount as $id => $count) {
            $name = sprintf("%-21s", $msgs[$id]);
            $col  = $arrcol[$msgs[$id]];
            $opts[] = "DEF:v$id=$libdir/$id.rrd:hits:AVERAGE"; // rrd datasrc 
            $opts[] = "CDEF:gv$id=v$id,UN,0,v$id,IF";          // make UNK=0
            $opts[] = "CDEF:gtvt$id=gv$id,$rate,*";            // height/rate 
            $opts[] = "$type:gtvt$id#$col:$name ";             // name&graph
            $opts[] = "CDEF:gvt$id=gv$id,$length,*";           // how many?
            $opts[] = "GPRINT:gvt$id:AVERAGE:%.0lf";           // how many?
            $opts[] = "COMMENT:\\n"; 
            $type = "STACK";
    }

    $opts[] = "COMMENT:amavis-stats v$asVersion\\r";

    $start = microtime();
    $ret = rrd_graph($img, $opts, count($opts));
    $t = elapsed($start);
    if (!is_array($ret)) {
        $err = rrd_error();
        asErr("rrd_graph(): $err");
        return false;
    }

    return "<img src=\"$me/$img\" alt=\"[image: $timetext]\">\n\n";
}


/*
* asVGraph($img, $end, $span)
*
* $img     - name of the png to generate
* $end     - end time in seconds
* $span    - string either "daily", "weekly", "monthly", or "yearly"
*
* Build a graph of Virus infected emails.
* Returns either a html-valid <img> tag which can be printed, or the
* boolean "false" if something went wrong.
*/
function asVGraph($img, $end, $span) {
    global $libdir, $me, $virus, $pid, $psid, $iid, $bid, $nid, $nsid;
    global $maxi, $rate, $asVersion, $rrdstep, $host;

    /*
    * Work out the beginning and the end, based on the span
    */
    $limits   = getlimits($end, $span);
    $end      = $limits["end"];
    $start    = $limits["start"];
    $length   = $end - $start;
    $timetext = $limits["description"];

    $startdate = strftime("%c", $start);
    $enddate   = strftime("%c %Z", $end);
    $nowdate   = strftime("%c", time());

    $hostname = "on $host ";
    if ($host == "") {
#        $hostname = eregi_replace("\n", "", `hostname`);
        $hostname = "";
    }


    /*
    * It is a two-step process to build the final graph. The average over
    * a specific time period seems to be impossible to get without actually
    * building a graph. Ie, rrd fetch will not calculate the values we
    * need - we would have to sum and average manually.
    *
    * However the PRINT function of a graph will return what we want
    * in an array. So first of all build a graph that PRINTs the average
    * of every virus over the selected time period.
    */

    $opts = array();
    $opts[] = "--start=$start";
    $opts[] = "--end=$end";

    foreach ($virus as $id => $rest) {
        $opts[] = "DEF:v$id=$libdir/$id.rrd:hits:AVERAGE";
        $opts[] = "CDEF:gv$id=v$id,UN,0,v$id,IF";
        $opts[] = "CDEF:gvt$id=gv$id,$length,*";
        $opts[] = "PRINT:gvt$id:AVERAGE:%.0lf";
    }

    $ret = rrd_graph($img, $opts, count($opts));

    /*
    * debugging - graph definitions
    */
    aslog($ret);

    $infected = 0;

    if (!is_array($ret)) {
        $msg = rrd_error();
        asErr("rrd_graph(): $msg");
        return false;
    }

    /*
    * All results from PRINT commands are in the array $ret["calcpr"][..]
    */
	$maxi = 0;
    $i = 0;
    $vcount = array();
    foreach ($virus as $id => $rest) {
        /*
        * We don't have enough resolution in the rrds
        * to calculate the correct counts at low averages,
        * so we just don't display them
        */
        if ($ret["calcpr"][$i] != 0) {
            $vcount[$id] = $ret["calcpr"][$i];
            $maxi = max($maxi, strlen($virus[$id]["name"]));
        }
        $i++;
        /* aslog("Id: $id = $vcount[$id]"); */
    }
    $maxi++;
    aslog("Maxi: $maxi");

    /*
    */
    if (count($vcount) >= 1) {
        arsort($vcount);
    }
    else {
        aslog("vcount is an empty array");
    }
    aslog($vcount);

    if ($rate == 3600) {
        $ratemsg = "hour";
    } else {
        $rate = 60;
        $ratemsg = "min";
    }


    /*
    * Now that we have the counts of each virus over the time period
    * we can build the actual graph
    */
    $opts = array();
    $opts[] = "--start=$start";
    $opts[] = "--end=$end";
    $opts[] = "--imgformat=PNG";
    $opts[] = "--title=Virus Detection $hostname($timetext to $enddate)";
    $opts[] = "--width=520";
    $opts[] = "--vertical-label=viruses/$ratemsg";

    /*
    * The tricky part, building rrd rows but ordering the elements by
    * columns...
    */

    if ($maxi > 20) {
        $width = 2;
    } else {
        $width = 3;
    }
    $total = count($vcount);
    $depth = ceil($total / $width);

    $mod = $total % $width;

    $infected =1;
    if ($infected > 0) {

        $keyarray = array_keys ($vcount);

        for ($d = 1; $d <= $depth; $d++) {
            for ($col = 1; $col <= $width; $col++) {

                if ($col == 1) {
                    $index = $d;
                }
                elseif ($d != $depth || $mod == 0 || $mod >= $col) {
                    if (($mod == 0) || ($col - $mod) < 2) {
                        $index = ($col - 1) * $depth + $d;
                    } else {
                        $index = $mod * $depth + 
                                 ($col - $mod - 1)*($depth - 1) + $d ;
                    }
                }
                else {
                    continue;
                }

                $id = $keyarray[$index - 1];

                if ($d == 1 && $col == 1) {
                    addopts($opts, "AREA", $id, $vcount, $virus, $length);
                }
                else {
                    addopts($opts, "STACK", $id, $vcount, $virus, $length);
                }
            }
            $opts[] = "COMMENT:\\n"; 
        }
    }

    $opts[] = "COMMENT:\\n"; 
    $opts[] = "COMMENT:amavis-stats v$asVersion\\r";

    aslog($opts);
    $ret = rrd_graph($img, $opts, count($opts));

    if (!is_array($ret)) {
        $err = rrd_error();
        asErr("rrd_graph(): $err");
        return false;
    }

    return "<img src=\"$me/$img\" alt=\"[image: $timetext]\">\n\n";
}


?>
